@@ -1,4 +1,8 @@
-"""Context for Intermediate Representation"""
+"""
+The context of a generator
+
+Includes the function, its representation and its usage information
+"""
 from __future__ import annotations
 
 import ast
@@ -9,9 +13,12 @@ from types import FunctionType
 from typing import Any, Optional
 
 from python2verilog.api.modes import Modes
-from python2verilog.ir.expressions import Var
+from python2verilog.ir.expressions import ExclusiveVar, State, Var
+from python2verilog.ir.graph import DoneNode
+from python2verilog.ir.instance import Instance
+from python2verilog.ir.signals import ProtocolSignals
 from python2verilog.utils.assertions import assert_typed_dict, get_typed, get_typed_list
-from python2verilog.utils.env_vars import is_debug_mode
+from python2verilog.utils.env import is_debug_mode
 from python2verilog.utils.generics import GenericReprAndStr
 
 DEFAULT_STATE_NAME = "___PYTHON_2_VERILOG_STATE___"
@@ -26,6 +33,8 @@ class Context(GenericReprAndStr):
 
     # pylint: disable=too-many-instance-attributes
     name: str = ""
+    testbench_suffix: str = "_tb"
+
     test_cases: list[tuple[int]] = field(default_factory=list)
 
     py_func: Optional[FunctionType] = None
@@ -38,31 +47,43 @@ class Context(GenericReprAndStr):
     optimization_level: int = -1
 
     mode: Modes = Modes.NO_WRITE
-    _module_file: Optional[io.IOBase] = None
-    _testbench_file: Optional[io.IOBase] = None
 
     _global_vars: list[Var] = field(default_factory=list)
     _input_vars: list[Var] = field(default_factory=list)
     _output_vars: list[Var] = field(default_factory=list)
     _states: set[str] = field(default_factory=set)
 
-    valid_signal: Var = Var("valid")
-    done_signal: Var = Var("done")
-    clock_signal: Var = Var("clock")
-    start_signal: Var = Var("start")
-    reset_signal: Var = Var("reset")
-    ready_signal: Var = Var("ready")
+    signals: ProtocolSignals = ProtocolSignals(
+        start=Var("start"),
+        done=Var("done"),
+        ready=Var("ready"),
+        valid=Var("valid"),
+    )
 
     state_var: Var = Var("state")
 
-    _entry_state: Optional[str] = None
-    _ready_state: Optional[str] = None
+    _done_state: State = State("_state_done")
+    idle_state: State = State("_state_idle")
+    _entry_state: Optional[State] = None
 
-    def __del__(self):
-        if self._module_file:
-            self._module_file.close()
-        if self._testbench_file:
-            self._testbench_file.close()
+    # Function calls
+    namespace: dict[str, Context] = field(default_factory=dict)  # callable functions
+    instances: dict[str, Instance] = field(default_factory=dict)  # generator instances
+
+    @property
+    def testbench_name(self) -> str:
+        """
+        Returns test bench module name in the generated verilog
+        """
+        return f"{self.name}{self.testbench_suffix}"
+
+    def _repr(self):
+        """
+        Avoids recursion on itself
+        """
+        dic = self.__dict__
+        del dic["namespace"]
+        return dic
 
     def validate(self):
         """
@@ -83,32 +104,20 @@ class Context(GenericReprAndStr):
         def check_list(list_: list):
             return isinstance(list_, list) and len(list_) > 0
 
-        assert check_list(self.input_types), self
+        assert check_list(self.input_types), "Input types not inferred"
         assert check_list(self.input_vars), self
 
         assert check_list(self.output_types), self
         assert check_list(self.output_vars), self
 
         assert isinstance(self.optimization_level, int), self
-        assert self.optimization_level >= 0
-
-        if self.mode:
-            if self._module_file:
-                assert isinstance(self.module_file, io.IOBase), self
-            if self._testbench_file:
-                assert isinstance(self.testbench_file, io.IOBase), self
+        assert self.optimization_level >= 0, f"{self.optimization_level} {self.name}"
 
         if self._entry_state:
-            assert self.entry_state in self.states, self
-        if self._ready_state:
-            assert self.ready_state in self.states, self
+            assert str(self.entry_state) in self.states, self
 
-        assert get_typed(self.ready_signal, Var)
-        assert get_typed(self.clock_signal, Var)
-        assert get_typed(self.done_signal, Var)
-        assert get_typed(self.valid_signal, Var)
-        assert get_typed(self.reset_signal, Var)
-        assert get_typed(self.start_signal, Var)
+        for value in self.signals.values():
+            assert get_typed(value, Var)
 
         return self
 
@@ -156,26 +165,26 @@ class Context(GenericReprAndStr):
         """
         The first state that does work in the graph representation
         """
-        assert isinstance(self._entry_state, str), self
+        assert isinstance(self._entry_state, State), self
         return self._entry_state
 
     @entry_state.setter
-    def entry_state(self, other: str):
-        assert isinstance(other, str)
+    def entry_state(self, other: State):
+        assert isinstance(other, State)
         self._entry_state = other
 
     @property
-    def ready_state(self):
+    def done_state(self):
         """
         The ready state
         """
-        assert isinstance(self._ready_state, str), self
-        return self._ready_state
+        assert isinstance(self._done_state, State), self
+        return self._done_state
 
-    @ready_state.setter
-    def ready_state(self, other: str):
-        assert isinstance(other, str)
-        self._ready_state = other
+    @done_state.setter
+    def done_state(self, other: State):
+        assert isinstance(other, State)
+        self._done_state = other
 
     @property
     def input_vars(self):
@@ -204,7 +213,7 @@ class Context(GenericReprAndStr):
         Sets own output vars to default based on number of output variables
         """
         assert self.output_types and len(self.output_types) > 0
-        self._output_vars = [Var(str(i)) for i in range(len(self.output_types))]
+        self._output_vars = [Var(f"out{i}") for i in range(len(self.output_types))]
 
     @property
     def global_vars(self):
@@ -221,6 +230,13 @@ class Context(GenericReprAndStr):
         """
         Appends global var
         """
+        var = get_typed(var, Var)
+        if (
+            var in self._global_vars
+            or var in self._input_vars
+            or var in self.output_vars
+        ):
+            return
         self._global_vars.append(get_typed(var, Var))
 
     @property
@@ -268,6 +284,7 @@ class Context(GenericReprAndStr):
     def __check_types(
         self, expected_types: list[type[Any]], actual_values: list[type[Any]]
     ):
+        assert len(expected_types) == len(actual_values)
         for expected, actual in zip(expected_types, actual_values):
             assert isinstance(
                 actual, expected
@@ -285,3 +302,34 @@ class Context(GenericReprAndStr):
         Checks if outputs to functions' types matches previous outputs
         """
         self.__check_types(self.output_types, output)
+
+    def create_instance(self, name: str) -> Instance:
+        """
+        Create generator instance
+        """
+        inst_input_vars: list[Var] = list(
+            map(
+                lambda var: ExclusiveVar(f"{name}_{self.name}_{var.py_name}"),
+                self.input_vars,
+            )
+        )
+        inst_output_vars: list[Var] = list(
+            map(
+                lambda var: ExclusiveVar(f"{name}_{self.name}_{var.py_name}"),
+                self.output_vars,
+            )
+        )
+        args: dict[str, Var] = {
+            key: ExclusiveVar(f"{name}_{self.name}__{value.py_name}")
+            for key, value in self.signals.instance_specific_items()
+        }
+
+        signals = ProtocolSignals(**args)
+
+        return Instance(
+            self.name,
+            Var(name),
+            inst_input_vars,
+            inst_output_vars,
+            signals,
+        )
