@@ -9,9 +9,9 @@ Element := Vertex | Edge
 
 from __future__ import annotations
 
-from typing import Optional
+from typing import Generator, Optional
 
-from ..utils.assertions import get_typed, get_typed_list
+from ..utils.assertions import get_typed, get_typed_list, get_typed_optional
 from . import expressions as expr
 
 
@@ -148,13 +148,13 @@ class BasicElement(Element):
         self._optimal_child = get_typed(other, Element)
 
 
-class Vertex(Element):
+class Node(Element):
     """
     Vertex
     """
 
 
-class IfElseNode(Vertex, Element):
+class IfElseNode(Node, Element):
     """
     Represents an if-else statement
     """
@@ -263,8 +263,28 @@ class IfElseNode(Vertex, Element):
         assert self._optimal_true_edge and self._optimal_false_edge
         return [self._optimal_true_edge, self._optimal_false_edge]
 
+    def traverse_condition_vars(self) -> Generator[expr.Var, None, None]:
+        """
+        Yields variables of the if conditional, doing a DPS
+        """
 
-class AssignNode(Vertex, BasicElement):
+        def rec(exp: expr.Expression):
+            if isinstance(exp, expr.UBinOp):
+                yield from rec(exp.left)
+                yield from rec(exp.right)
+            elif isinstance(exp, expr.UnaryOp):
+                yield from rec(exp.expr)
+            elif isinstance(exp, expr.Var):
+                yield exp
+            elif isinstance(exp, (expr.UInt, expr.Int)):
+                pass
+            else:
+                raise RuntimeError(f"{type(exp)}")
+
+        yield from rec(self.condition)
+
+
+class AssignNode(Node, BasicElement):
     """
     Represents a non-blocking assignment,
     i.e. assignments that do not block the execution of
@@ -308,8 +328,14 @@ class AssignNode(Vertex, BasicElement):
         """
         return f"{self._lvalue.to_string()} <= {self._rvalue.to_string()}"
 
+    def verilog(self):
+        """
+        To string
+        """
+        return f"{self._lvalue.verilog()} <= {self._rvalue.verilog()}"
 
-class YieldNode(Vertex, BasicElement):
+
+class YieldNode(Node, BasicElement):
     """
     Yield statement, represents output
     """
@@ -335,10 +361,14 @@ class YieldNode(Vertex, BasicElement):
         """
         To string
         """
-        return f"yield {str(self._stmts)}"
+        string = "yield ["
+        for stmt in self._stmts:
+            string += stmt.to_string() + ", "
+        string = string[:-2] + "]"
+        return string
 
 
-class DoneNode(Vertex, Element):
+class DoneNode(Node, Element):
     """
     Signals done
     """
